#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include "dalloc.h"
#include "mpi-utils.h"

#define NDIMS 2

static inline double
sum_prod (double * a, double * b, size_t n) {
  size_t i;
  double s = 0.0;

  for (i = 0; i < n; i++)
    s += a[i]*b[i];

  return s;
}

int main (int argc, char * argv[]) {
  /* general */
  int i, j;

  int m = 5*13;
  int n = 7*11;

  /* MPI specific */
  int mpi_size;
  int mpi_rank;

  MPI_Comm mpi_comm_cart;
  MPI_Comm mpi_comm_cols;
  MPI_Comm mpi_comm_rows;

  /* local */
  int local_m, local_n;
  int start_m, start_n;

  double ** A;
  double  * b;
  double  * c;

  int   dims[NDIMS] = {0};
  int coords[NDIMS] = {0};

  /* root */
  double * root_b = NULL;
  double * root_c = NULL;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);

  /* take user input */
  if (argc > 2) {
    m = atoi(argv[1]);
    n = atoi(argv[2]);
  }

  /* get suggestion on best work divison */
  MPI_Dims_create(mpi_size, NDIMS, dims);

  /* create a new communicator with the
     layout of a cartesian grid */
  MPI_Cart_create(MPI_COMM_WORLD, NDIMS,
		  dims, (int[NDIMS]){0},
		  false, &mpi_comm_cart);

  /* get local coordinates on grid */
  MPI_Cart_coords(mpi_comm_cart, mpi_rank, NDIMS, coords);

  /* make communicators that work across
     entire rows or columns */
  MPI_Comm_split(MPI_COMM_WORLD, coords[0], coords[1],
		 &mpi_comm_cols);
  MPI_Comm_split(MPI_COMM_WORLD, coords[1], coords[0],
		 &mpi_comm_rows);

  /* calculate slices */
  start_m = BLOCK_LOW (coords[1], dims[1], m);
  local_m = BLOCK_SIZE(coords[1], dims[1], m);

  start_n = BLOCK_LOW (coords[0], dims[0], n);
  local_n = BLOCK_SIZE(coords[0], dims[0], n);

  /* root initializes b and c */
  if (PROCESS_IS_ROOT(mpi_rank)) {
    root_b = malloc(n * sizeof(double));
    root_c = malloc(m * sizeof(double));

    if (root_b == NULL || root_c == NULL) {
      fprintf(stderr,
	      "process %d: allocation failed",
	      mpi_rank);
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }

    for (j = 0; j < n; j++)
      root_b[j] = (double) j/(n-1);
  }

  /* allocate and initialize
     local block matrix */
  A = dalloc(local_m, local_n, sizeof(double));

  if (A == NULL) {
    fprintf(stderr,
	    "process %d: allocation failed",
	    mpi_rank);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  for (i = 0; i < local_m; i++)
    for (j = 0; j < local_n; j++)
      A[i][j] = (double)
      	((i + start_m) * n + (j + start_n))/(m*n-1);

  /* everyone allocates local slices */
  b = malloc(local_n * sizeof(double));
  c = malloc(local_m * sizeof(double));

  if (b == NULL || c == NULL) {
    fprintf(stderr,
	    "process %d: allocation failed\n",
	    mpi_rank);
    MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
  }

  /* root shares b with every
     process in the first row */
  if (PROCESS_IS_ROOT(coords[1])) {
    int * counts, * displs;

    counts = malloc(dims[0] * sizeof(int));
    displs = malloc(dims[0] * sizeof(int));

    if (counts == NULL || displs == NULL) {
      fprintf(stderr,
	      "process %d: allocation failed\n",
	      mpi_rank);
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }

    for (i = 0; i < dims[0]; i++) {
      counts[i] = BLOCK_SIZE(i, dims[0], n);
      displs[i] = BLOCK_LOW (i, dims[0], n);
    }

    MPI_Scatterv(root_b, counts, displs, MPI_DOUBLE,
		 b, local_n, MPI_DOUBLE,
		 ROOT_PROCESS, mpi_comm_rows);

    free(displs);
    free(counts);
  }

  /* every process in the first row
     shares its share of b with every
     process in its column */
  MPI_Bcast(b, local_n, MPI_DOUBLE, ROOT_PROCESS,
	    mpi_comm_cols);

  /* local subdomain matrix
     vector multiplication */
  /* NB */
  for (i = 0; i < local_m; i++)
    c[i] = sum_prod(A[i], b, local_n);

  /* Every row reduces its share of c */
  MPI_Reduce(PROCESS_IS_ROOT(coords[0]) ? MPI_IN_PLACE : c,
	     c, local_m, MPI_DOUBLE, MPI_SUM,
	     ROOT_PROCESS, mpi_comm_rows);

  /* root gathers the complete c
     vector from every process in
     the first column */
  if (PROCESS_IS_ROOT(coords[0])) {
    int * counts, * displs;

    counts = malloc(dims[1] * sizeof(int));
    displs = malloc(dims[1] * sizeof(int));

    if (counts == NULL || displs == NULL) {
      fprintf(stderr,
	      "process %d: allocation failed\n",
	      mpi_rank);
      MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE);
    }

    for (i = 0; i < dims[1]; i++) {
      counts[i] = BLOCK_SIZE(i, dims[1], m);
      displs[i] = BLOCK_LOW (i, dims[1], m);
    }

    MPI_Gatherv(c, local_m, MPI_DOUBLE,
		root_c, counts, displs, MPI_DOUBLE,
		ROOT_PROCESS, mpi_comm_cols);

    free(displs);
    free(counts);
  }

  /* root prints */
  if (PROCESS_IS_ROOT(mpi_rank))
    for (i = 0; i < m; i++)
      printf("%d: %f\n", i, root_c[i]);

  /* cleanup and deallocate */
  free(c);
  free(b);
  dalloc_free(A);

  free(root_c);
  free(root_b);

  MPI_Comm_free(&mpi_comm_rows);
  MPI_Comm_free(&mpi_comm_cols);
  MPI_Comm_free(&mpi_comm_cart);

  FINALIZE_AND_SUCCEED;
}
