#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ndalloc.h"
#include "mpi-utils.h"

#define NDIMS 2

static void block_mult(double ** C, double ** A, double ** B, size_t n) {
  size_t i, j, k;

  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      for (k = 0; k < n; k++)
	C[i][j] += A[i][k] * B[k][j];
}

int main (int argc, char * argv[]) {
  /* general */
  int i, j, k;

  int n = 4;

  /* MPI specific */
  int mpi_size;
  int mpi_rank;

  MPI_Comm mpi_comm_cart;
  MPI_Comm mpi_comm_cols;
  MPI_Comm mpi_comm_rows;

  /* local */
  int local_n;
  int start_m;
  int start_n;

  double ** block_A;
  double ** block_B;
  double ** block_C;

  double *** slice_A;
  double *** slice_B;

  int   dims[NDIMS] = {0};
  int coords[NDIMS] = {0};

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);

  /* take user input */
  if (argc > 1) {
    n = atoi(argv[1]);
  }

  /* get suggestion on best work divison */
  MPI_Dims_create(mpi_size, NDIMS, dims);

  if (dims[0] != dims[1]) {
    root_eprintf(mpi_rank, "%s: not a square number of processes\n",
		 argv[0]);
    FINALIZE_AND_FAIL;
  }

  if (n % mpi_size) {
    root_eprintf(mpi_rank, "%s: perfect work division not possible\n",
		 argv[0]);
    FINALIZE_AND_FAIL;
  }

  /* create a new communicator with the
     layout of a cartesian grid */
  MPI_Cart_create(MPI_COMM_WORLD, NDIMS,
		  dims, (int[NDIMS]){0},
		  false, &mpi_comm_cart);

  /* get local coordinates on grid */
  MPI_Cart_coords(mpi_comm_cart, mpi_rank, NDIMS, coords);

  /* make communicators that work across
     entire rows or columns */
  MPI_Comm_split(MPI_COMM_WORLD, coords[1], coords[0],
		 &mpi_comm_cols);
  MPI_Comm_split(MPI_COMM_WORLD, coords[0], coords[1],
		 &mpi_comm_rows);

  /* calculate slices */
  local_n = n/dims[0];

  start_m = BLOCK_LOW(coords[0], dims[0], n);
  start_n = BLOCK_LOW(coords[1], dims[1], n);

  /* allocate and initialize
     local block matrices */
  block_A = ndalloc(2, sizeof(double), local_n, local_n);
  block_B = ndalloc(2, sizeof(double), local_n, local_n);
  block_C = ndalloc(2, sizeof(double), local_n, local_n);

  for (i = 0; i < local_n; i++) {
    for (j = 0; j < local_n; j++) {
      block_A[i][j] = (i + start_m) * n + j + start_n;
      block_B[i][j] = mpi_rank;
    }
  }

  /* allocate and initialize
     local matrix slices. */
  slice_A = ndalloc(3, sizeof(double), 3, local_n, local_n);
  slice_B = ndalloc(3, sizeof(double), 3, local_n, local_n);

  /* Allgather instead of Allgatherv since we assume
     work is perfectly divisible. */
  /* everyone shares blocks of A to form a slice
     in the row direction. */
  MPI_Allgather(block_A[0],    local_n * local_n, MPI_DOUBLE,
		slice_A[0][0], local_n * local_n, MPI_DOUBLE,
		mpi_comm_rows);

  /* everyone shares blocks of B to form a slice
     in the column direction. */
  MPI_Allgather(block_B[0],    local_n * local_n, MPI_DOUBLE,
		slice_B[0][0], local_n * local_n, MPI_DOUBLE,
		mpi_comm_cols);

  /* set every element of C to zero. */
  memset(&block_C[0][0], 0, local_n*local_n * sizeof(double));
  for (k = 0; k < dims[0]; k++)
    block_mult(block_C, slice_A[k], slice_B[k], local_n);

  /* synchronized printout. */
  {
    int synch = 1;

    MPI_Recv(&synch, 1, MPI_INT,
	     mpi_rank ? mpi_rank-1 : MPI_PROC_NULL,
	     101, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

    printf("%d (%d, %d):\n", mpi_rank, coords[0], coords[1]);
    puts("A:");
    for (i = 0; i < local_n; i++) {
      for (j = 0; j < local_n; j++)
	printf(" %f (%f)", slice_A[coords[1]][i][j], block_A[i][j]);

      puts("");
    }

    puts("B:");
    for (i = 0; i < local_n; i++) {
      for (j = 0; j < local_n; j++)
	printf(" %f (%f)", slice_B[coords[0]][i][j], block_B[i][j]);

      puts("");
    }

    puts("C:");
    for (i = 0; i < local_n; i++) {
      for (j = 0; j < local_n; j++)
	printf(" %f", block_C[i][j]);

      puts("");
    }
    fflush(stdout);

    MPI_Send(&synch, 1, MPI_INT,
	     mpi_rank < mpi_size-1 ? mpi_rank+1 : MPI_PROC_NULL,
	     101, MPI_COMM_WORLD);
  }

  ndalloc_free(3, slice_B);
  ndalloc_free(3, slice_A);

  ndalloc_free(2, block_C);
  ndalloc_free(2, block_B);
  ndalloc_free(2, block_A);

  MPI_Comm_free(&mpi_comm_rows);
  MPI_Comm_free(&mpi_comm_cols);
  MPI_Comm_free(&mpi_comm_cart);

  FINALIZE_AND_SUCCEED;
}
